/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

/* Copyright 2014, 2015 Endless Mobile, Inc. */

/*
 * This file is part of eos-event-recorder-daemon.
 *
 * eos-event-recorder-daemon is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or (at your
 * option) any later version.
 *
 * eos-event-recorder-daemon is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with eos-event-recorder-daemon.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "config.h"
#include <string.h>

#include <gio/gio.h>
#include <glib.h>
#include <glib-object.h>
#include <glib-unix.h>
#include <polkit/polkit.h>

#include "emer-daemon.h"
#include "emer-event-recorder-server.h"
#include "shared/metrics-util.h"

typedef struct _DBusCallbackData
{
  EmerEventRecorderServer *server;
  GDBusMethodInvocation *invocation;
} DBusCallbackData;

static gboolean
on_record_singular_event (EmerEventRecorderServer *server,
                          GDBusMethodInvocation   *invocation,
                          guint32                  user_id,
                          GVariant                *event_id,
                          gint64                   relative_timestamp,
                          gboolean                 has_payload,
                          GVariant                *payload,
                          EmerDaemon              *daemon)
{
  emer_daemon_record_singular_event (daemon, event_id, relative_timestamp,
                                     has_payload, payload);
  emer_event_recorder_server_complete_record_singular_event (server,
                                                             invocation);
  return TRUE;
}

static gboolean
on_record_aggregate_event (EmerEventRecorderServer *server,
                           GDBusMethodInvocation   *invocation,
                           guint32                  user_id,
                           GVariant                *event_id,
                           gint64                   count,
                           gint64                   relative_timestamp,
                           gboolean                 has_payload,
                           GVariant                *payload,
                           EmerDaemon              *daemon)
{
  emer_event_recorder_server_complete_record_aggregate_event (server,
                                                              invocation);
  return TRUE;
}

static gboolean
on_record_event_sequence (EmerEventRecorderServer *server,
                          GDBusMethodInvocation   *invocation,
                          guint32                  user_id,
                          GVariant                *event_id,
                          GVariant                *events,
                          EmerDaemon              *daemon)
{
  emer_daemon_record_event_sequence (daemon, user_id, event_id, events);
  emer_event_recorder_server_complete_record_event_sequence (server,
                                                             invocation);
  return TRUE;
}

static gboolean
on_set_enabled (EmerEventRecorderServer *server,
                GDBusMethodInvocation   *invocation,
                gboolean                 enabled,
                EmerDaemon              *daemon)
{
  EmerPermissionsProvider *permissions =
    emer_daemon_get_permissions_provider (daemon);

  emer_permissions_provider_set_daemon_enabled (permissions, enabled);
  emer_permissions_provider_set_uploading_enabled (permissions, enabled);
  emer_event_recorder_server_complete_set_enabled (server, invocation);
  return TRUE;
}

static gboolean
on_start_aggregate_timer (EmerEventRecorderServer *object,
                          GDBusMethodInvocation   *invocation,
                          guint32                  unix_user_id,
                          GVariant                *event_id,
                          gboolean                 has_payload,
                          GVariant                *payload,
                          EmerDaemon              *daemon)
{
  g_autofree gchar *timer_object_path = NULL;
  g_autoptr(GError) error = NULL;
  GDBusConnection *system_bus;
  const gchar *sender;

  system_bus = g_dbus_method_invocation_get_connection (invocation);
  sender = g_dbus_method_invocation_get_sender (invocation);

  emer_daemon_start_aggregate_timer (daemon,
                                     system_bus,
                                     sender,
                                     unix_user_id,
                                     event_id,
                                     has_payload,
                                     payload,
                                     &timer_object_path,
                                     &error);

  if (error)
    {
      g_dbus_method_invocation_return_gerror (invocation, error);
      return TRUE;
    }

  emer_event_recorder_server_complete_start_aggregate_timer (object,
                                                             invocation,
                                                             timer_object_path);

  return TRUE;
}

static void
handle_upload_finished (EmerDaemon       *daemon,
                        GAsyncResult     *result,
                        DBusCallbackData *callback_data)
{
  GError *error = NULL;
  gboolean upload_succeeded =
    emer_daemon_upload_events_finish (daemon, result, &error);
  if (upload_succeeded)
    emer_event_recorder_server_complete_upload_events (callback_data->server,
                                                       callback_data->invocation);
  else
    g_dbus_method_invocation_take_error (callback_data->invocation, error);

  g_object_unref (callback_data->server);
  g_object_unref (callback_data->invocation);
  g_free (callback_data);
}

static gboolean
on_upload_events (EmerEventRecorderServer *server,
                  GDBusMethodInvocation   *invocation,
                  EmerDaemon              *daemon)
{
  DBusCallbackData *callback_data = g_new (DBusCallbackData, 1);
  callback_data->server = g_object_ref (server);
  callback_data->invocation = g_object_ref (invocation);
  emer_daemon_upload_events (daemon,
                             (GAsyncReadyCallback) handle_upload_finished,
                             callback_data);
  return TRUE;
}

typedef struct {
    const gchar *method_name;
    const gchar *method_full_name;
    const gchar *error_message;
} AuthorizedMethod;

static const AuthorizedMethod authorized_methods[] = {
   {
     "SetEnabled",
     "com.endlessm.Metrics.SetEnabled",
     "Disabling metrics is only allowed from system settings"
   },
   {
     "ResetTrackingId",
     "com.endlessm.Metrics.ResetTrackingId",
     "Only privileged users can reset the tracking ID"
   },
   { NULL }
};

static const AuthorizedMethod *
lookup_authorized_method (const gchar *method_name)
{
  gsize i;

  for (i = 0; i < G_N_ELEMENTS (authorized_methods); ++i)
    {
      if (g_strcmp0 (authorized_methods[i].method_name, method_name) == 0)
        return &authorized_methods[i];
    }

  return NULL;
}

/*
 * This handler is run in a separate thread, so all operations can be
 * synchronous.
 */
static gboolean
on_authorize_method_check (GDBusInterfaceSkeleton *interface,
                           GDBusMethodInvocation  *invocation,
                           EmerDaemon             *daemon)
{
  const gchar *method_name =
    g_dbus_method_invocation_get_method_name (invocation);
  const AuthorizedMethod *authorized_method = lookup_authorized_method (method_name);

  if (authorized_method == NULL)
    return TRUE;

  GError *error = NULL;
  PolkitAuthority *authority =
    polkit_authority_get_sync (NULL /*GCancellable*/, &error);
  if (authority == NULL)
    {
      g_critical ("Could not get PolicyKit authority: %s.", error->message);
      g_dbus_method_invocation_return_gerror (invocation, error);
      g_error_free (error);
      return FALSE;
    }

  const gchar *sender_name = g_dbus_method_invocation_get_sender (invocation);
  PolkitSubject *subject = polkit_system_bus_name_new (sender_name);
  PolkitCheckAuthorizationFlags flags = POLKIT_CHECK_AUTHORIZATION_FLAGS_NONE;
  GDBusMessage *message = g_dbus_method_invocation_get_message (invocation);

  if (g_dbus_message_get_flags (message) & G_DBUS_MESSAGE_FLAGS_ALLOW_INTERACTIVE_AUTHORIZATION)
    flags |= POLKIT_CHECK_AUTHORIZATION_FLAGS_ALLOW_USER_INTERACTION;

  PolkitAuthorizationResult *result =
    polkit_authority_check_authorization_sync (authority,
                                               subject,
                                               authorized_method->method_full_name,
                                               NULL /*PolkitDetails*/,
                                               flags,
                                               NULL /*GCancellable*/,
                                               &error);
  g_object_unref (authority);
  g_object_unref (subject);
  if (result == NULL)
    {
      g_critical ("Could not get PolicyKit authorization result: %s.",
                  error->message);
      g_dbus_method_invocation_return_gerror (invocation, error);
      g_error_free (error);
      return FALSE;
    }

  gboolean authorized = polkit_authorization_result_get_is_authorized (result);
  if (!authorized)
    g_dbus_method_invocation_return_error_literal (invocation,
                                                   G_DBUS_ERROR,
                                                   G_DBUS_ERROR_AUTH_FAILED,
                                                   authorized_method->error_message);

  g_object_unref (result);
  return authorized;
}

typedef struct {
  EmerDaemon *daemon;
  GMainLoop  *main_loop;
} ShutdownSignalData;

static gboolean
quit_main_loop (ShutdownSignalData *data)
{
  g_assert (data != NULL);
  g_assert (EMER_IS_DAEMON (data->daemon));
  g_assert (data->main_loop != NULL);

  emer_daemon_shutdown (data->daemon);
  g_main_loop_quit (data->main_loop);
  return G_SOURCE_REMOVE;
}

static gboolean
sync_recorder_server_enabled_for_upload (GBinding     *binding,
                                         const GValue *from_value,
                                         GValue       *to_value,
                                         gpointer      user_data)
{
  g_autoptr(EmerPermissionsProvider) permissions =
    EMER_PERMISSIONS_PROVIDER (g_binding_dup_source (binding));

  gboolean daemon_enabled =
    emer_permissions_provider_get_daemon_enabled (permissions);
  gboolean uploading_enabled = g_value_get_boolean (from_value);

  g_value_set_boolean (to_value, daemon_enabled && uploading_enabled);

  return TRUE;
}

static gboolean
sync_recorder_server_enabled_for_daemon (GBinding     *binding,
                                         const GValue *from_value,
                                         GValue       *to_value,
                                         gpointer      user_data)
{
  g_autoptr(EmerPermissionsProvider) permissions =
    EMER_PERMISSIONS_PROVIDER (g_binding_dup_source (binding));

  gboolean daemon_enabled = g_value_get_boolean (from_value);
  gboolean uploading_enabled =
    emer_permissions_provider_get_uploading_enabled (permissions);

  g_value_set_boolean (to_value, daemon_enabled && uploading_enabled);

  return TRUE;
}

/*
 * Called when a reference to the system bus is acquired. This is where you are
 * supposed to export your well-known name, not in name_acquired; that is too
 * late.
 */
static void
on_bus_acquired (GDBusConnection *system_bus,
                 const gchar     *name,
                 gpointer         user_data)
{
  EmerDaemon *daemon = EMER_DAEMON (user_data);
  EmerEventRecorderServer *server = emer_event_recorder_server_skeleton_new ();

  g_signal_connect (server, "handle-record-singular-event",
                    G_CALLBACK (on_record_singular_event), daemon);
  g_signal_connect (server, "handle-record-aggregate-event",
                    G_CALLBACK (on_record_aggregate_event), daemon);
  g_signal_connect (server, "handle-record-event-sequence",
                    G_CALLBACK (on_record_event_sequence), daemon);
  g_signal_connect (server, "handle-set-enabled",
                    G_CALLBACK (on_set_enabled), daemon);
  g_signal_connect (server, "handle-upload-events",
                    G_CALLBACK (on_upload_events), daemon);
  g_signal_connect (server, "handle-start-aggregate-timer",
                    G_CALLBACK (on_start_aggregate_timer), daemon);
  g_signal_connect (server, "g-authorize-method",
                    G_CALLBACK (on_authorize_method_check), daemon);

  EmerPermissionsProvider *permissions =
    emer_daemon_get_permissions_provider (daemon);
  g_object_bind_property_full (permissions, "daemon-enabled", server, "enabled",
                               G_BINDING_DEFAULT | G_BINDING_SYNC_CREATE,
                               sync_recorder_server_enabled_for_daemon, NULL,
                               daemon, NULL);
  g_object_bind_property_full (permissions, "uploading-enabled", server, "enabled",
                               G_BINDING_DEFAULT | G_BINDING_SYNC_CREATE,
                               sync_recorder_server_enabled_for_upload, NULL,
                               daemon, NULL);

  GError *error = NULL;
  if (!g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON (server),
                                         system_bus,
                                         "/com/endlessm/Metrics",
                                         &error))
    g_error ("Could not export metrics interface on system bus: %s.",
             error->message);
}

/*
 * This is called if ownership of the well-known name is lost. Since this
 * service doesn't own and un-own the name during its lifetime, this is only
 * called if there is an error acquiring it in the first place.
 */
static void
on_name_lost (GDBusConnection *system_bus,
              const gchar     *name,
              gpointer         user_data)
{
  /*
   * This handler is called with a NULL connection if the bus could not be
   * acquired.
   */
  if (system_bus == NULL)
    g_error ("Could not get connection to system bus.");

  g_error ("Could not acquire name '%s' on system bus.", name);
}

static EmerDaemon *
make_daemon (gint                argc,
             const gchar * const argv[])
{
  g_autofree gchar *persistent_cache_directory = NULL;
  const gchar *persistent_cache_directory_const = PERSISTENT_CACHE_DIR;
  g_autofree gchar *config_file_path = NULL;
  g_autoptr(EmerPermissionsProvider) permissions_provider = NULL;
  GOptionEntry option_entries[] =
  {
    { "persistent-cache-directory", 'p', G_OPTION_FLAG_NONE,
      G_OPTION_ARG_FILENAME, &persistent_cache_directory,
      "Store persistent cache at path", "path"},
    { "config-file-path", 'c', G_OPTION_FLAG_NONE,
      G_OPTION_ARG_FILENAME, &config_file_path,
      "Path to permissions config file", "path"},
    { NULL }
  };

  GOptionContext *option_context =
    g_option_context_new (NULL /* parameter string */);
  g_option_context_add_main_entries (option_context, option_entries,
                                     NULL /* translation domain */);

  GError *error = NULL;
  gboolean parse_succeeded =
    g_option_context_parse (option_context, &argc, (gchar ***) &argv, &error);
  g_option_context_free (option_context);

  if (!parse_succeeded)
    {
      g_warning ("Option parsing failed: %s.", error->message);
      g_error_free (error);
      return NULL;
    }

  if (persistent_cache_directory != NULL)
    persistent_cache_directory_const = persistent_cache_directory;

  if (config_file_path != NULL)
    permissions_provider =
      emer_permissions_provider_new_full (config_file_path, NULL);

  return emer_daemon_new (persistent_cache_directory_const,
                          permissions_provider);
}

gint
main (gint                argc,
      const gchar * const argv[])
{
  EmerDaemon *daemon = make_daemon (argc, argv);
  if (daemon == NULL)
    return EXIT_FAILURE;

  GMainLoop *main_loop = g_main_loop_new (NULL, TRUE);
  ShutdownSignalData data = { daemon, main_loop };

  // Shut down on any of these signals.
  g_unix_signal_add (SIGHUP, (GSourceFunc) quit_main_loop, &data);
  g_unix_signal_add (SIGINT, (GSourceFunc) quit_main_loop, &data);
  g_unix_signal_add (SIGTERM, (GSourceFunc) quit_main_loop, &data);
  g_unix_signal_add (SIGUSR1, (GSourceFunc) quit_main_loop, &data);
  g_unix_signal_add (SIGUSR2, (GSourceFunc) quit_main_loop, &data);

  guint name_id = g_bus_own_name (G_BUS_TYPE_SYSTEM, "com.endlessm.Metrics",
                                  G_BUS_NAME_OWNER_FLAGS_NONE,
                                  on_bus_acquired,
                                  NULL /* name_acquired_callback */,
                                  on_name_lost,
                                  daemon, NULL /* user data free func */);

  g_main_loop_run (main_loop);

  g_object_unref (daemon);
  g_bus_unown_name (name_id);
  g_main_loop_unref (main_loop);

  return EXIT_SUCCESS;
}
